// math.h only use for data initialization

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cfft16.h"

#define DBG_CFFT16

int turn(cfft_t *dst, cfft_t *src, int row, int col);

// cfft 16 init 
int cfft16_init(cfft16plan_t *cfft16planHandle){    
    int i, j;
    int N = 16;
    double PI = atan(1)*4;
    cfft_t *twiddle = NULL;    
    
    // cfft4 data init
    cfft16planHandle->cfft4planHandle = NULL;
    cfft16planHandle->cfft4planHandle = malloc(sizeof(cfft4plan_t));
    cfft4_init(cfft16planHandle->cfft4planHandle);
    
    cfft16planHandle->inputBuf = malloc(sizeof(cfft_t)*16);
    cfft16planHandle->turnBuf  = malloc(sizeof(cfft_t)*16);
    cfft16planHandle->twiddle  = malloc(sizeof(cfft_t)*12);
    twiddle = cfft16planHandle->twiddle;    
    //twiddle factor init, initial code neednot care computing efficency    
    for(i = 0; i < 3; i ++){//row loop
      for(j = 0; j < 4; j ++){
        twiddle[i*4 + j].r = cos(-2*PI/N*(i+1)*j);
        twiddle[i*4 + j].i = sin(-2*PI/N*(i+1)*j);
      }
    }
#ifdef DBG_CFFT16    
    ComplexPrint("cfft16Twiddle", twiddle, 12);
#endif    
    return 0;        
}
// in place fft support
int cfft16_proc(
      cfft16plan_t *cfft16planHandle,
      cfft_t *in,
      cfft_t *out)
{
    int i;
    cfft_t multTemp;    // complex multiplication buffer
    cfft_t *inputBuf, *turnBuf, *twiddle;
    cfft4plan_t *cfft4planHandle = cfft16planHandle->cfft4planHandle;   
    
    turnBuf     = cfft16planHandle->turnBuf;
    inputBuf    = cfft16planHandle->inputBuf;
    twiddle     = cfft16planHandle->twiddle;
    
#ifdef DBG_CFFT16    
    ComplexPrint("cfft16In", in, 16);
#endif    
    // map the input by col order
    turn(inputBuf, in, 4, 4);    
#ifdef DBG_CFFT16    
    ComplexPrint("afterTurn", inputBuf, 16);
#endif    
    // column FFT
    for(i = 0; i < 16; i += 4){      
      cfft4_proc(cfft4planHandle, &(inputBuf[i]), &(inputBuf[i]));
    }
#ifdef DBG_CFFT16    
    ComplexPrint("ColFFT", inputBuf, 16);
#endif    
    turn(turnBuf, inputBuf, 4, 4);
#ifdef DBG_CFFT16    
    ComplexPrint("beforeTW", turnBuf, 16);
#endif    
    // twiddle 
    for(i = 4; i < 16; i ++){
      multTemp.r   = turnBuf[i].r; multTemp.i = turnBuf[i].i;      
      turnBuf[i].r = twiddle[i-4].r*multTemp.r - twiddle[i-4].i*multTemp.i;
      turnBuf[i].i = twiddle[i-4].r*multTemp.i + twiddle[i-4].i*multTemp.r;
    }
#ifdef DBG_CFFT16    
    ComplexPrint("afterTW", turnBuf, 16);    
#endif    
    for(i = 0; i < 16; i += 4){      
      cfft4_proc(cfft4planHandle, &(turnBuf[i]), &(turnBuf[i]));
    }
    turn(out, turnBuf, 4, 4);
#ifdef DBG_CFFT16    
    ComplexPrint("out", out, 16);
#endif    
    return 0;
}

int cfft16_close(cfft16plan_t *cfft16planHandle){    
    cfft4_close(cfft16planHandle->cfft4planHandle);
    free(cfft16planHandle->cfft4planHandle);
    free(cfft16planHandle->inputBuf);
    free(cfft16planHandle->turnBuf);
    free(cfft16planHandle->twiddle);
    return 0;
}

// cfft 4 init 
int cfft4_init(cfft4plan_t *cfft4planHandle){    
    cfft4planHandle->inputBuf = malloc(sizeof(cfft_t)*4);
    return 0;    
}

// in place fft support
int cfft4_proc(
      cfft4plan_t *cfft4planHandle,
      cfft_t *in,
      cfft_t *out)
{
    cfft_t *inputBuf;
        
    inputBuf = cfft4planHandle->inputBuf;
    
    memcpy(inputBuf, in, sizeof(cfft_t)*4);// compiler const substitute
    
    out[0].r=inputBuf[0].r+inputBuf[1].r+inputBuf[2].r+inputBuf[3].r;
    out[0].i=inputBuf[0].i+inputBuf[1].i+inputBuf[2].i+inputBuf[3].i;
    out[1].r=inputBuf[0].r+inputBuf[1].i-inputBuf[2].r-inputBuf[3].i;
    out[1].i=inputBuf[0].i-inputBuf[1].r-inputBuf[2].i+inputBuf[3].r;
    out[2].r=inputBuf[0].r-inputBuf[1].r+inputBuf[2].r-inputBuf[3].r;
    out[2].i=inputBuf[0].i-inputBuf[1].i+inputBuf[2].i-inputBuf[3].i;
    out[3].r=inputBuf[0].r-inputBuf[1].i-inputBuf[2].r+inputBuf[3].i;
    out[3].i=inputBuf[0].i+inputBuf[1].r-inputBuf[2].i-inputBuf[3].r;
    
    return 0;
}

int cfft4_close(cfft4plan_t *cfft4planHandle){        
    free(cfft4planHandle->inputBuf);
    return 0;
}

void ComplexPrint(char *name, cfft_t *data, int N){
    int i;
    fprintf(stdout, "%s = [\n", name);
	for(i = 0; i < N; i ++){
	  fprintf(stdout,"%+f%+fj\n", data[i].r, data[i].i);
	}
	fprintf(stdout,"]\n");
}
// virtual 2d array turn, donot support in place turn
turn(cfft_t *dst, cfft_t *src, int srcRow, int srcCol){
    int i = 0, j = 0, dIndex = 0, sIndex = 0;    
    if(dst == src){
      fprintf(stdout, "error, turn(dst,src,....), dst MUST != src\n ");
      return 1;
    }
    // by the row of source array   
    for(i = 0; i < srcRow; i ++){
      dIndex = i;
      // by the column of source array
      for(j = 0; j < srcCol; j ++){
        dst[dIndex].r = src[sIndex].r;
        dst[dIndex].i = src[sIndex].i;
        dIndex += srcRow;   // src Row num is the dst Col num
        sIndex += 1;
      }
    }
    return 0;
}
